#include "VRML Viewpoints.h"
#define RESOLUTION WORLD_RESOLUTION
#define AEZERO(x) (fabs(x)<RESOLUTION) 
#define AEZERO2(x) (fabs(x)<WORLD_RESOLUTION) 
#define FastSetInRangeForD(theta,min,increment)\
			(\
				(FORD)((min)+(increment)*(((theta)-(min))/(increment)-floor(((theta)-(min))/(increment))))\
			)

void MakeMatrixForRotation(FORD (*Mat)[3],FORD theta, FORD ax,FORD by,FORD cz)
{
	FORD a=ax,b=by,c=cz;
	FORD fcos_theta=fcos(theta);
	FORD fsin_theta=fsin(theta);
	Mat[0][0]=a*a*(1-fcos_theta)+fcos_theta;
	Mat[0][1]=a*b*(1-fcos_theta)-c*fsin_theta;
	Mat[0][2]=a*c*(1-fcos_theta)+b*fsin_theta;
	Mat[1][0]=a*b*(1-fcos_theta)+c*fsin_theta;
	Mat[1][1]=b*b*(1-fcos_theta)+fcos_theta;
	Mat[1][2]=b*c*(1-fcos_theta)-a*fsin_theta;
	Mat[2][0]=a*c*(1-fcos_theta)-b*fsin_theta;
	Mat[2][1]=b*c*(1-fcos_theta)+a*fsin_theta;
	Mat[2][2]=c*c*(1-fcos_theta)+fcos_theta;
}

void MatMult(FORD (*Mat1)[3],FORD (*Mat2)[3],FORD (*Result)[3])
{
	int i,j;
	for(i=0;i<3;i++)
	for(j=0;j<3;j++)
	{
		int k;
		Result[i][j]=0;
		for(k=0;k<3;k++)
			Result[i][j]+=Mat1[i][k]*Mat2[k][j];
	}	
}

void FindLengthAndAngleFrom1stAxisAsForD(FORD *ptheta,FORD *pradius,double x,double y)
{
	double ratio=x/(*pradius);
	ratio=min(1,max(-1,ratio));//in case rounding error takes its abs value over 1
	*pradius=(FORD)max(1e-10,(float)sqrt(x*x+y*y));
	*ptheta=(FORD)acos((double)x/(*pradius));
	if(*pradius==0)
		*ptheta=0;
	if(y<0)*ptheta=-*ptheta;
	*ptheta=FastSetInRangeForD(*ptheta,-PI,TWO_PI);
}

BOOL is_solution(FORD (*a)[3],FORD *x)
{
 int i,j;
 for(i=0;i<3;i++)
 {
  double sum=0;
  for(j=0;j<3;j++)
   sum+=a[i][j]*x[j];
  if((fabs(sum)>=RESOLUTION))
   return FALSE;
 }
 return TRUE;
}

char solve_sim_eqs_for_non_zero_sol(FORD (*a)[3],FORD *x)
                                   //coefficients, solution
//solves system of 3 simultaneous equations
//arguments should be 
//double a[3][3]; ==coefficients
//double x[3]; == solution
//RESOLUTION is absolute value below which entries are to be regarded as zero
//Each row of coefficients has the coefficients for one of the equations.
//The equations are
//a[j][0]*x[0] +a[j][1]*x[1]+a[j][2]*x[2]==0
//looks for solution with at least one non zero value of x[].
//method is, try eliminating 0th, then 1st, then 2nd coefficient. For each one, try eliminating
//it using 1st, 2nd, or 3rd equation as the one to subtract
{
 int i,j,row_to_subtract,ith,jth;
 double A[3][3];//scratch array for calcs - reset from a[][] each time
 //check to see if default axis is a solution - can happen e.g. if all entries in r are zero
	x[0]=x[2]=0;x[1]=1;
 if(is_solution(a,x))
  return 1;
 for(row_to_subtract=0;row_to_subtract<3;row_to_subtract++)
 for(ith=0;ith<3;ith++)
 {
  if(fabs(a[row_to_subtract][ith])>=RESOLUTION)
  {
   //reset scratch array ready to start new elimination
   for(i=0;i<3;i++)
   for(j=0;j<3;j++)
    A[j][i]=a[j][i];
   //eliminate ith coefficient from all other rows with ith coefficient non zero
   for(jth=0;jth<3;jth++)
   if(jth!=row_to_subtract)
   {
    //find multiplier 
    double mki=A[jth][ith]/A[row_to_subtract][ith];//so mki*A[row_to_subtract][ith]==A[jth][ith]
    //subtract row_to_subtract*multiplier from jth row
    if(mki!=0)
    for(i=0;i<3;i++)
    {
     A[jth][i]-=mki*A[row_to_subtract][i];
     if(fabs(A[jth][i])<RESOLUTION)
      A[jth][i]=0;
    }
   }
   //now A[j][] has ith coord zero for both of the j!=row_to_subtract
   //find non zero solution to transformed equations, if any
   {
    int i1=(ith+1)%3,i2=(ith+2)%3,i3=ith;                    
    int j1=(row_to_subtract+1)%3,j2=(row_to_subtract+2)%3,j3=row_to_subtract;
    //The i3==ith coefficients were eliminated from rows j1 and j2
    //so rows j1 and j2 have to be identical up to a factor for a non zero solution of the transformed equations
    //so can use either to find x[i1] and x[i2]
    //use A[j2][i1]*x[i1]+A[j2][i2]*x[i2]==0 to find x[i1] and x[i2]
    if(fabs(A[j2][i2])>=RESOLUTION)
    {
     x[i1]=1;
     x[i2]=-A[j2][i1]/A[j2][i2];
    }
    else 
     if(fabs(A[j2][i1])>=RESOLUTION)
     {
      x[i2]=1;
      x[i1]=-A[j2][i2]/A[j2][i1];
     }
     else
      //only solution to the transformed equations is x[i1]==x[i2]==x[i3]==0
      continue;
    //use A[j3][i1]*x[i1]+A[j3][i2]*x[i2]+A[j3][i3]*x[i3]==0 to find x[i3]
    if(fabs(A[j3][i3])>=RESOLUTION)
     x[i3]=-(A[j3][i1]*x[i1]+A[j3][i2]*x[i2])/A[j3][i3];
    else continue;//only solution to the transformed equations is x[i1]==x[i2]==x[i3]==0
    //check it is indeed a solution
			 if(is_solution(a,x))
					return 1;
   }
  }
 }
 x[0]=x[2]=0;x[1]=1;//may as well return default axis - shouldn't happen
 return 0;
}

#ifdef cl_up_vector_discards
void cross_product(double *A,double *B,double *C)
{
 /**formulae
 Cx=AyBz-AzBy
 Cy=AzBx-AxBz
 Cz=AxBy-AyBx
 **/
 C[0]=A[1]*B[2]-A[2]*B[1];
 C[1]=A[2]*B[0]-A[0]*B[2];
 C[2]=A[0]*B[1]-A[1]*B[0];
}
#endif

void MakeViewpoint(TCHAR *szViewpoint)
{
	//szViewpoint should be of size MAX_SZ_VIEWPOINT + enough chars for rest of viewpoint
	FORD Mat1[3][3],Mat2[3][3],Mat3[3][3];
	FORD Result[3][3];
	FORD r[3][3];
	FORD new_axis[3];
	FORD theta=0;
	FORD diff[3];
	int i;
	FORD th_about_z,th_about_y,th_about_x;
	FORD radius;
	FORD diffn[3],ldiffn=0;
	char ok_sol=0;
	//normalised diff - direction cosines
	for(i=0;i<WORLD_DIMENSION;i++)
		diff[i]=view_to[i]-view_from[i];
	for(i=0;i<WORLD_DIMENSION;i++)
		ldiffn+=diff[i]*diff[i];
	if(ldiffn<1e-15)
		//view_to and view_from are identical - use default
	{ldiffn=1;diff[0]=diff[1]=0;diff[1]=1;}
	for(i=0;i<WORLD_DIMENSION;i++)
	{
		diffn[i]=diff[i]/(FORD)sqrt(ldiffn);
  diffn[i]=min(1,max(-1,diffn[i]));//in case rounding error takes its abs value over 1
	}
	//need to apply tilt first, then angle about x, then finally, about y
 th_about_z=view_tilt;
	//angle about x is PI/2 + angle from vertical (y) (yields -ve Z axis if all angles 0)
	//have ndiff[1]==cos(angle from vertical)
	th_about_x=(FORD)asin(diffn[1]);
	th_about_x=FastSetInRangeForD((FORD)th_about_x,-PI,TWO_PI);
	//angle about y is angle about vertical (y)
	FindLengthAndAngleFrom1stAxisAsForD(&th_about_y,&radius,-diffn[2],-diffn[0]);
	if(AEZERO2(radius))
		th_about_y=0;
  if(th_about_z!=0)
   MakeMatrixForRotation(Mat1,th_about_z,0,0,1);
	MakeMatrixForRotation(Mat2,th_about_x,1,0,0);
	MakeMatrixForRotation(Mat3,th_about_y,0,1,0);
 //Find Mat3.Mat2.Mat1 - leave out Mat1 if theta_about_z==0
 if(th_about_z==0)
  MatMult(Mat3,Mat2,r);
 else
 {
  MatMult(Mat2,Mat1,Result);
  MatMult(Mat3,Result,r);
 }
	//Now need to find axis of rotation for result
	//want eigenvector for (r-I)x=0
	//where r==rotation array
	r[0][0]-=1;
	r[1][1]-=1;
	r[2][2]-=1;
	//want to solve
	//r[0][0]*x[0]+r[0][1]*x[1]+r[0][2]*x[2]==0
	//r[1][0]*x[0]+r[1][1]*x[1]+r[1][2]*x[2]==0
	//r[2][0]*x[0]+r[2][1]*x[1]+r[2][2]*x[2]==0
	//where x==new_axis

	ok_sol=solve_sim_eqs_for_non_zero_sol(r,new_axis);
	//now need to find the rotation about that axis
 if(ok_sol>0)
	{
  //convert new_axis to direction cosines
		FORD a=new_axis[0],b=new_axis[1],c=new_axis[2];
		FORD length=(FORD)sqrt(a*a+b*b+c*c);
		a/=length;
		b/=length;
		c/=length;
		new_axis[0]=a;new_axis[1]=b;new_axis[2]=c;
		for(i=0;i<WORLD_DIMENSION;i++)
			if(fabs(new_axis[i])<WORLD_RESOLUTION)
				new_axis[i]=0;
  //Now need new angle
 {
   //Solve r.(0,0,-1)=diffn where r is rotation about new axis
   //so need to look at 3rd col of the rotation matrix for rotation
   //by theta about new axis
   //and want
			//a*c*(1-cos(theta))+b*sin(theta)==-diffn[0]
			//b*c*(1-cos(theta))-a*sin(theta)==-diffn[1]
			//c*c*(1-cos(theta))+cos(theta)==-diffn[2]
			//so
			//(1-c*c)cos(theta)==-diffn[2]-(c*c);
			if(AEZERO2(1-c*c))
				//rotation about z axis
			{new_axis[0]=new_axis[1]=0;new_axis[2]=1;theta=FastSetInRangeForD(view_tilt,-PI,TWO_PI);}
			else
			{
				double arg=(-diffn[2]-c*c)/(1-c*c);
				arg=min(1,max(-1,arg));//in case rounding error takes its abs value over 1
				theta=(FORD)acos(arg);
				theta=FastSetInRangeForD(theta,-PI,TWO_PI);
				{
					FORD fcos_theta=fcos(theta),fsin_theta=fsin(theta);
					FORD t1=a*c*(1-fcos_theta)+b*fsin_theta;//ti should be -Result2[i][2] to reconstruct original matrix using theta
					FORD t2=b*c*(1-fcos_theta)-a*fsin_theta;
					FORD t3=c*c*(1-fcos_theta)+fcos_theta;
					{
						if(
							!AEZERO2(t1+diffn[0])
							||!AEZERO2(t2+diffn[1])
							||!AEZERO2(t3+diffn[2])
							)
							theta=TWO_PI-theta;
						theta=FastSetInRangeForD(theta,-PI,TWO_PI);
					}
				}
			}
		}
	}
	if(fabs(theta)<RESOLUTION)
		theta=0;
	{
  char sztilt[100];
  sztilt[0]='\0';
  if(view_tilt)
  sprintf(sztilt," tilt %.*g %s",precision,view_tilt*angles_mult
         ,SZ_ANGLES_IN
         );
		sprintf
   (szViewpoint,
"Viewpoint\n"
"{\n description \"%s\"\n" // ,szViewpointName
" position %.*g %.*g %.*g\n" // ,precision,view_from[0],precision,view_from[1],precision,view_from[2],precision,theta
" orientation %.*g %.*g %.*g %.*g # towards %.*g %.*g %.*g %s%s\n" 
   //,precision,new_axis[0],precision,new_axis[1],precision,new_axis[2]  ,precision,view_to[0],precision,view_to[1],precision,view_to[2]
" fieldOfView %.*g\n"// ,precision,view_field_of_view
 "}\n"
		 ,szViewpointName
   ,precision,view_from[0],precision,view_from[1],precision,view_from[2]
   ,precision,new_axis[0],precision,new_axis[1],precision,new_axis[2],precision,theta
   ,precision,view_to[0],precision,view_to[1],precision,view_to[2]
   ,sztilt
		 ,ok_sol?"":"!! no solution found for rotation axis - using default\n"
   ,precision,view_field_of_view
		 );
	}
}
