#include "VRML Viewpoints.h"
#pragma hdrstop( "" )
#define IS_CLOSE_BRACKET(c)  (c==')'||c==']'||c=='}')
#define IS_OPEN_BRACKET(c)  (c=='('||c=='['||c=='{')

void SetDlgEditTextWithCaretToEnd(HWND hDlg,WORD idd,PSTR szEdit)
{
 char Buffer[STRMAX];
 int strlen_szEdit=max(0,strlen(szEdit));
 HWND hwndEdit=GetDlgItem(hDlg,idd);
 GetDlgItemText(hDlg,idd,Buffer,STRMAX);
 if(strcmp(Buffer,szEdit)!=0)
 {
  WORD was_idc_mid_update=idc_mid_update;
  idc_mid_update=idd;
  SetDlgItemText(hDlg,idd,szEdit);
  idc_mid_update=was_idc_mid_update;
  UpdateWindow(hwndEdit);
  SendMessage(hwndEdit,EM_SETSEL,strlen_szEdit,strlen_szEdit);
  UpdateWindow(hwndEdit);
 }
 else 
  SendMessage(hwndEdit,WM_SETREDRAW,TRUE,0L);
}

long SendMessageForID(HWND hwnd,WORD id,UINT message,WPARAM wParam,LPARAM lParam)
{
	long lreturn;
	HWND hwnd_id=hwnd?GetDlgItem(hwnd,id):0;
	if(hwnd)lreturn=SendMessage(hwnd_id,message,wParam,lParam);
	return lreturn;
}

int next_token_pos_for_white_space(char *sz,int *pireturn)
/*
goes through the string like strtok,call with sz = string to search 
each time, and call with *pireturn<0 first time.
Sets *pireturn to start of next token as coord rather than pointer. 
Doesn't add any '\0's to separate the tokens.
Sets ipos to -1 for end of string.
Returns 1, or 0 for end of string.
*/
{
int ipos=*pireturn;
if(ipos<0)
 ipos=0;
else
 {
 /*find next separator*/
 for(;;ipos++)
  {
  char c=sz[ipos];
  switch(c)
   {
   case '\0':
   CASE_ISSPACE
   goto found_separator;
   break;
   }
  }
 }
/*ready for next token search*/
found_separator:;
/*find next non separator*/
for(;;ipos++)
 {
 char c=sz[ipos];
 if(c=='\0')
  {
  *pireturn=-1;
  return 0;
  }
 switch(c)
  {
  CASE_ISSPACE
   break;
  default:
   *pireturn=ipos;
   return 1;
		}
 }
return 1;
}

int strcatsprintf(UINT *puPos,UINT uMax,PSTR pstr,const TCHAR *format,...)
/*
example of use:
strcatsprintf(&uPos,szText,format,...);
Initialise iPos to 0 before first call.
The iPos is kept up to date by this proc, so that the next call places
Returns without doing 
the new text at the end of the old one.
anything if the new text would increase pstr to more than uMax characters.

Faster than strcat() for arrays etc as it adds new strings to current position reached instead of
requiring strcat to re-scan szText from the start to the null terminator each time.

Increases speed by a factor of order n! where n is the number of strings to be concatenated by it.
*/
{
TCHAR message[MAX_SZEDIT];
TCHAR *pArguments=(TCHAR *)&format+sizeof format;
int nchars_written=vsprintf(message,format,pArguments);
if(pstr==NULL)
 return 0;
if(*puPos+nchars_written>=uMax)
 return 0;
strcpy(&pstr[*puPos],message); 
(*puPos)+=nchars_written;
return nchars_written;
}

long double NZtanl(long double theta)
/*
version to use for quotients to avoid divide by 0 errors, so
program can continue
*/
/*this one is used by slong_double*/
{
long double sin_theta=sinl(theta);
long double cos_theta=cosl(theta);
if(sin_theta==0.0)
 {
 return 0.0000000000000000001;
 }
if(cos_theta==0.0)
 {
 return 100000000000000000.0;
 }
else
 return sin_theta/cos_theta;
}

char is_numeral(char *string)
{
	int i;
	for(i=0;;i++)
	{
		char c=string[i];
		if(c=='\0')
			return 0;
		if(isspace(c))
			return 0;
		switch(c)
		{
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':return 1;
		case '.':
		case '-':
		case '+':break;
		default:return 0;
		}
	}
	return 0;
}

//new: non_white_space_found + is_numeral(..)
void ParseExpression(const char *format,char *string,long double *pLd
		    ,float *pf,int *psh,char *pparsed)
{
 /*
 Expression parser assuming no white space
 This means e.g. (1+2) 3 is read as two numbers, while (1+2)3 is read as a
 single number
 Could remove all white space within brackets before calling it.
 *pparsed is set to one if any of the atoms are evaluated.
 */
 char ft2=format[1];
 long double ld=*pLd,ld1,ld2;
 float f=*pf,f1,f2;
 int sh=*psh,sh1,sh2;
 int i,strlen_string=strlen(string);      
 int nbrackets=0,level;
 double max_power=38;
 char non_white_space_found=!isspace(string[0]);
 int next_sign=1;
 char found_next_sign_prev=0;
 switch(ft2)
 {             
 case 'L':max_power=307;break;
 case 'd':max_power=4.5;break;
 }
#define LD1 ((ft2=='L'?ld1:((ft2=='g'||ft2=='f')?f1:sh1)))
#define LD2 (ft2=='L'?ld2:((ft2=='g'||ft2=='f')?f2:sh2))
#define MAKE_RESULT(result)                     \
 switch(ft2)                                \
 {                                          \
   case 'L':ld=(long double)(result);break; \
   case 'g':                                \
   case 'f':f=(float)(result);break;        \
   case 'd':sh=(int)(result);break;         \
 }
 if(!non_white_space_found)
 {
	 MAKE_RESULT(0) goto to_return;
 }
 if(string[0]=='(')
 /*if entire expression is in brackets, can remove them*/
 for(i=0;string[i];i++)
 {
  char c=string[i];
  if(IS_OPEN_BRACKET(c))
   nbrackets++;
  if(IS_CLOSE_BRACKET(c))
   nbrackets--;
  if(nbrackets==0)
  {
   char cp1=string[i+1];
   if(isspace(cp1)||cp1=='\0')
   {
    if(c==')')
    {
     string[i]='\0';
     ParseExpression(format,&(string[1]),&ld,&f,&sh,pparsed);
     string[i]=c;
     goto to_return;
    }
   }
   else
    break;
  }
 }
 for(level=0;level<=2;level++)
 for(i=0,nbrackets=0;string[i];i++)
 {
  char c=string[i],cm1=string[max(i-1,0)];
  char oldc=c;
  int lb_multiplication_offset=1;
  if(isspace(c))
   break;
  if(i>0&&level==1&&nbrackets==0)
  if((c=='('&&cm1!='c'&&cm1!='C'&&cm1!='e'&&cm1!='E'&&cm1!='l'&&cm1!='L'&&cm1!='n'&&cm1!='N'
     &&cm1!='t'&&cm1!='T'&&cm1!='r'&&cm1!='R'&&cm1!='s'&&cm1!='S'&&cm1!='+'&&cm1!='-'&&cm1!='*'&&cm1!='/'&&cm1!='|'&&cm1!='^'
     )
    ||(cm1==')'&&c!='+'&&c!='-'&&c!='*'&&c!='/'&&c!='|'&&c!='^')
    )
  {/*treat as multiplication for 5(3+2) etc (3+2)5 etc*/ 
   /*but not for c(..), e(..) etc*/
   c='*';
   lb_multiplication_offset=0;
  }
  if(c=='('&&lb_multiplication_offset!=0)
   nbrackets++;
  if(c==')')
   nbrackets--;
  if(found_next_sign_prev||(i==0&&level==0))
  {
   switch(c)
    {
     case '-':// unary minus
      if(IS_OPEN_BRACKET(string[i+1]))
      {
       ParseExpression(format,&string[i+1],&ld1,&f1,&sh1,pparsed);
       MAKE_RESULT(-LD1)
	      goto to_return;
      } 
      else
      {
       next_sign=-next_sign;
       found_next_sign_prev=1;
      }
      break;
     default:
      if(found_next_sign_prev)
      {
       ParseExpression(format,&string[i],&ld1,&f1,&sh1,pparsed);
       MAKE_RESULT(next_sign*LD1)
	      goto to_return;
      } 
      found_next_sign_prev=0;
    }
  }
  if(nbrackets==0&&i>0)
  switch(level)
  {
  case 0:
   switch(c)
    {
    case '+':
    case '-':
     if(cm1=='e'||cm1=='E')goto end_loop;/*expressions such as 1e-10*/
     if(cm1=='*'||cm1=='/'||cm1=='-'||cm1=='+')goto end_loop;/*expressions such as *- or +- or -- etc*/
     string[i]='\0';
     ParseExpression(format,string,&ld1,&f1,&sh1,pparsed);
     ParseExpression(format,&(string[i+1]),&ld2,&f2,&sh2,pparsed);
     string[i]=oldc;
    }
   switch(c)
    {
    case '+':MAKE_RESULT(LD1+LD2);goto to_return;
    case '-':MAKE_RESULT(LD1-LD2);goto to_return;
    }
   break;
  case 1:
   switch(c)
    {
    case '*':if(string[i+1]=='*'){i++;goto end_loop;}
    case '/':
     string[i]='\0';
     ParseExpression(format,string,&ld1,&f1,&sh1,pparsed);
     string[i]=oldc;
     ParseExpression(format,&(string[i+lb_multiplication_offset]),&ld2,&f2,&sh2,pparsed);
    }
   switch(c)
    {
    case '*':MAKE_RESULT(LD1*LD2);goto to_return;
    case '/':if(LD2==0)MAKE_RESULT(LD1)
	     else MAKE_RESULT(LD1/LD2)
	     goto to_return;
    }
   break;
  case 2:
   switch(c)
    {
    case '*':if(string[i+1]!='*')goto end_loop;
    case '|':/* using 2**3 or 2|3 or 2^3 for 2 to the power of 3*/
    case '^':
     string[i]='\0';
     ParseExpression(format,string,&ld1,&f1,&sh1,pparsed);
     ParseExpression(format,&(string[i+(c=='*'?2:1)]),&ld2,&f2,&sh2,pparsed);
     string[i]=oldc;
    }
   switch(c)
    {
    case '*':
    case '|':
    case '^':
     if(LD2*log((double)LD1)>max_power*log((double)10)
       ||LD2*log((double)LD1)<-max_power*log((double)10)
       )
      {
      MAKE_RESULT(LD1)
      goto to_return;
      }
     if(LD1==0)
      MAKE_RESULT(0)
     else
      if(LD2==0)
       MAKE_RESULT(1)
      else
       if(LD1<0)
							{
							int n=(int)LD2;
							MAKE_RESULT((long double)pow((double)LD1,(double)n)); 
							}
       else
								MAKE_RESULT((long double)pow((double)LD1,(double)LD2));
    goto to_return;
    }
   break;
  }
  end_loop:;
 }
	/*
	check for abbreviations of r5 for root 5 etc. Can be used with 
	brackets as r(expression)
	*/      
 {
  long double arg=0;
  switch(string[0])
  {
   case 'p':
   case 'P':MAKE_RESULT(next_sign*D_PI);*pparsed=1;goto to_return;
   case 'r':
   case 'R':
   case 'e':
   case 'E':
   case 'l':
   case 'L':
   case 'n':
   case 'N':
			 ParseExpression(format,&(string[1]),&ld1,&f1,&sh1,pparsed);
			 break;
   case 't':
   case 'T':
   case 'c':
   case 'C':
   case 's':
   case 'S':
			 ParseExpression(format,&(string[1]),&ld1,&f1,&sh1,pparsed);
			 arg=LD1;
  }
  switch(string[0])
  {
   case 'e':
   case 'E':
			 MAKE_RESULT(exp((double)LD1));
			 goto to_return;
   case 'l':
   case 'L':
			 MAKE_RESULT(log((double)LD1));goto to_return;
   case 'r':
   case 'R':
			 /*r5 for sqrtl(5), 1/r5 for 1/sqrtl(5)*/
			 MAKE_RESULT(sqrtl(LD1));goto to_return;
   case 't':
   case 'T':
			 /*T5 for tanl(5)*/
			 MAKE_RESULT(NZtanl((long double)arg));/*tanl and tan seem to calculate -sin(arg)/cos(arg)*/
			 goto to_return;
   case 'c':
   case 'C':
			 MAKE_RESULT(cosl((long double)arg));goto to_return;
   case 's':
   case 'S':
			 MAKE_RESULT(sinl((long double)arg));goto to_return;
  }
 }
 if(is_numeral(string))
	switch(format[1])
	{
		case 'L':if(sscanf(string,format,&ld)==1)*pparsed=1;break;
		case 'g':
		case 'f':if(sscanf(string,format,&f)==1)*pparsed=1;break;
		case 'd':if(sscanf(string,format,&sh)==1)*pparsed=1;break;
	}
	to_return:
	switch(format[1])
	{
		case 'L':*pLd=ld;break;
		case 'g':
		case 'f':*pf=f;break;
		case 'd':*psh=sh;break;
	}
}

void CompactifyFormula(char *szformula,char *szdisplay_formula)
{
	int i,ipos=0;
	for(i=0;;i++)
	{
		if(szdisplay_formula[i]=='\0')
			break;
		switch(szdisplay_formula[i])
		{
			CASE_ISSPACE
				continue;
		}
		szformula[ipos++]=szdisplay_formula[i];
		switch(szdisplay_formula[i])
		{
			case 'c':
			case 'C':if(substrcmpUpper(&szdisplay_formula[i],"COS")==0)i+=2;break;
			case 'e':
			case 'E':if(substrcmpUpper(&szdisplay_formula[i],"EXP")==0)i+=2;break;
			case 'l':
			case 'L':if(substrcmpUpper(&szdisplay_formula[i],"LOG")==0)i+=2;break;
			case 's':
			case 'S':
				if(substrcmpUpper(&szdisplay_formula[i],"SIN")==0)
					i+=2;
				else
				if(substrcmpUpper(&szdisplay_formula[i],"SQRT")==0||substrcmpUpper(&szdisplay_formula[i],"ROOT")==0)
				{szformula[ipos-1]='r';i+=3;}
				break;
			case 'p':
			case 'P':if(substrcmpUpper(&szdisplay_formula[i],"PI")==0)i++;break;
			case 't':
			case 'T':if(substrcmpUpper(&szdisplay_formula[i],"TAN")==0)i+=2;break;
		}
	}
	szformula[ipos]='\0';
}

void remove_space_in_brackets(PSTR string)
{
int i,nbrackets=0;
for(i=0;string[i];i++)
 {
 char c=string[i];
 if(IS_OPEN_BRACKET(c))
  nbrackets++;
 if(IS_CLOSE_BRACKET(c))
  nbrackets--;
 if(nbrackets>0)
 if(isspace(c))
  strcpy(&string[i],&string[i+1]);
 }
}

long double slong_double(char * string,long double dReturnValueForNULLOrUnParsedString)
{
	int itoken=-1;
	long double ld;float f;int sh;
	char parsed=0;
	if(string!=NULL)
	{
		//CompactifyFormula(string,string);
		remove_space_in_brackets(string);
		if(next_token_pos_for_white_space(string,&itoken))
		{
			ParseExpression("%Lg",&(string[itoken]),&ld,&f,&sh,&parsed);
			if(parsed)
				return ld;
		}
	}
	return dReturnValueForNULLOrUnParsedString;
}

long double slong_double_if_parsed(char * string,char *pparsed)
{
	int itoken=-1;
	long double ld;float f;int sh;
	if(string!=NULL)
	{
		if(next_token_pos_for_white_space(string,&itoken))
		{
			ParseExpression("%Lg",&(string[itoken]),&ld,&f,&sh,pparsed);
			if(pparsed)
				return ld;
		}
	}
	return 0;
}

char substrcmpUpper(const TCHAR * pstr1,const TCHAR * pstr2)
//returns 0 if the same, or pstr2 a sub string of pstr1
{
long i;
TCHAR c1[2],c2[2];
c1[1]=c2[1]='\0';
if(pstr1==NULL||pstr2==NULL)                          
 return 1;
for(i=0;;i++)
 {
 c1[0]=pstr1[i];
 c2[0]=pstr2[i];
 if(c2[0]=='\0')
	 return 0;
 CharUpper(c1);
 CharUpper(c2);
 if(c1[0]!=c2[0])
  return 1;
 }
}

char substrcmpUpperToWhiteSpace(const TCHAR * pstr1,const TCHAR * pstr2)
//returns 0 if the same, or pstr2 the same as pstr1 as far as next white space for both
{
long i;
TCHAR c1[2],c2[2];
c1[1]=c2[1]='\0';
if(pstr1==NULL||pstr2==NULL)                          
 return 1;
for(i=0;;i++)
 {
	char ok1=0,ok2=0;
 c1[0]=pstr1[i];
 c2[0]=pstr2[i];
 CharUpper(c1);
 CharUpper(c2);
	ok1=c1[0]=='\0'||isspace(c1[0]);
	ok2=c2[0]=='\0'||isspace(c2[0]);
	if(ok1&&ok2)
	 return 0;
	if(ok1||ok2||c1[0]!=c2[0])
  return 1;
 }
}

int MBsprintEcho(HWND hwnd,char *caption,UINT MB_s,char *format,...)
/*
works like sprintf, but to an MBEcho message box
MB_s is the 0 for OK button, etc.
MBsprintEcho(hwnd,caption,0,format,...);
*/
{
char message[2*STRMAX];
char *pArguments=(char *)&format+sizeof format;
vsprintf(message,format,pArguments);
return MessageBox(hwnd,message,caption,MB_s);
}

int copy_token_to_return(TCHAR *lptoken,TCHAR *lp,int ipos,int imax_copy)
/*
returns number of characters copied
ipos=pos to copy them from in lp 
*/
{
int i=0;
if(lp==NULL||lptoken==NULL||ipos<0) 
 return 0;
 /*find next separator*/
 for(;i<imax_copy;ipos++)
  {
  char c=lp[ipos];
  if(c=='\0'||c=='\n'||c=='\r')
   break;
  lptoken[i++]=c;
  }
lptoken[i]='\0';
return i;
}

void UnpackReturns(char *sz2,char *sz,DWORD max_length)
/*
translates returns to \015\012 s
sz == source
sz2== destination
*/
{
	char ch_return[]="\015\012";
	DWORD pos1,pos2/*,lstrlen_sz=lstrlen(sz)*/;
/*,lstrlen_sz2=lstrlen(sz2);*/
for(pos1=0,pos2=0;/*pos1<lstrlen_sz*/pos2<max_length;)
 {
 char c=sz[pos1++];
 /**
 if(pos2+3>=lstrlen_sz2) Is this meant to be memory size for sz2?
  break;
 **/
 if(c=='\0')
  break;
 if(c==ch_return[1]/*'\n'*/)
  {
  sz2[pos2++]=ch_return[0];
  sz2[pos2++]=ch_return[1];
  continue;
  }
 if(c==ch_return[0]/*'\r'*/)
  continue;
 if(c=='\013')
  continue;
 sz2[pos2++]=c;
 }
sz2[pos2]='\0';
return;
}
